.386p
.MODEL FLAT

DMA_BUF_MASK	EQU 8192-1 ; AUDIO_BUF_SIZE-1

	INCLUDE "apaula.i"

	; amixer.asm
	EXTERN _Mixer_MixVoices:PROC

	; areplay.asm
	EXTERN _Replayer_Tick:PROC

	; replayer.c
	EXTERN _Replayer_SongPlaying:BYTE

	; audio.c
	EXTERN _lpFilterTab:DWORD
	EXTERN _lastFilterSampleL:WORD
	EXTERN _lastFilterSampleR:WORD
	EXTERN _mixBufferL:DWORD
	EXTERN _mixBufferR:DWORD
	EXTERN _samplesPerTickHi:DWORD
	EXTERN _samplesPerTickLo:DWORD
	EXTERN _sampleCounterHi:DWORD
	EXTERN _sampleCounterLo:DWORD
	EXTERN _audioSideFactorMul:DWORD
	EXTERN _stereoSeparation:BYTE
	EXTERN _audio16Bit:BYTE
	EXTERN _stereoAudio:BYTE
	EXTERN _audioAmiga500:BYTE
	PUBLIC _Audio_MixSamplesDummy
	PUBLIC _Audio_MixSamples

.CODE

; -----------------------------------------------------------------------------
; LOW-PASS FILTER MACRO (A500 AUDIO MODE)
; -----------------------------------------------------------------------------

	;  AX = last 16-bit output sample
	; EDI = low-pass filter table
	; ESI = audio samples to filter
	;
	; NOTE: High-word of EAX and EBX must be cleared!
	;
FILTER_M MACRO pos
	sub ax,[edi+eax*2]
	mov bx,[esi+(pos*2)]
	add ax,[edi+ebx*2]
	mov [esi+(pos*2)],ax
ENDM

; -----------------------------------------------------------------------------
; DMA TRANSFER MACROS
; -----------------------------------------------------------------------------

	; 16-bit stereo
DMA16S_M MACRO pos
	xor eax,eax
	movsx ebx,word ptr [esi+(pos*2)]	; -16384..16256 (two Amiga voices)
	mov [esi+(pos*2)],ax			; clear L mix sample
	movsx edx,word ptr [ecx+(pos*2)]	; -16384..16256 (two Amiga voices)
	mov [ecx+(pos*2)],ax			; clear R mix sample
	mov eax,ebx
	add ebx,edx
	sub eax,edx
	imul esp
	sar eax,7
	mov edx,ebx
	add edx,eax			; edx = outL (-32768..32767 normalized)
	sub ebx,eax			; ebx = outR (-32768..32767 normalized)
	mov [edi+(pos*4)+0],bx		; output L
	mov [edi+(pos*4)+2],dx		; output R
ENDM

	; 16-bit stereo at center (mono mix, stereo output)
DMA16M_M MACRO pos
	mov dx,[esi+(pos*2)]
	mov [edi+(pos*4)+0],dx	; output L
	mov [edi+(pos*4)+2],dx	; output R
	mov [esi+(pos*2)],bx	; clear mix sample
ENDM

	; 8-bit mono (unsigned)
DMA8M_M MACRO pos ; four samples at once
	mov bh,[esi+(pos*8)+7]
	mov bl,[esi+(pos*8)+5]
	shl ebx,16
	mov bh,[esi+(pos*8)+3]
	mov bl,[esi+(pos*8)+1]
	xor ebx,edx		; convert samples to unsigned
	mov [esi+(pos*8)+0],esp	; clear what we read
	mov [esi+(pos*8)+4],esp
	mov [edi+(pos*4)],ebx
ENDM

	; Input:
	;  EAX = Number of samples to "mix" (silence)
	;
	; NOTE: This should only be used in apitint.asm.
	;       Thus, no registers are preserved on exit.
	;
	; CAUTION: Use DS segment override on base-EBP indexing!
	;
	ALIGN 4
_Audio_MixSamplesDummy PROC NEAR
	cmp _samplesPerTickHi,0
	je MixDone0
	; --------------------------------------------
SampleLoop0:
	test eax,eax ; samples left?
	jle MixDone0
	; --------------------------------------------
	; Check if it's time to tick the replayer...
	; --------------------------------------------
	mov edx,_sampleCounterHi
	test edx,edx
	jnz NoNewTick0
	cmp _Replayer_SongPlaying,1
	jne NoPlay0
	call _Replayer_Tick
NoPlay0:
	mov edx,_samplesPerTickHi
	mov ebx,_samplesPerTickLo
	add _sampleCounterLo,ebx
	adc edx,0
	mov _sampleCounterHi,edx
NoNewTick0:
	mov ecx,eax
	cmp ecx,edx
	jle Skip0 ; samplesLeft > samplesPerTick?
	mov ecx,edx
Skip0:	sub _sampleCounterHi,ecx
	sub eax,ecx
	jmp SampleLoop0
MixDone0:
	retn
_Audio_MixSamplesDummy ENDP

	; Input:
	;  EDI = Pointer to DMA buffer
	;  EAX = DMA buffer position
	;  EBX = Number of samples to mix
	;
	; NOTE: This should only be used in apitint.asm.
	;       Thus, no registers are preserved on exit.
	;
	; CAUTION: Use DS segment override on base-EBP indexing!
	;
	ALIGN 4
_Audio_MixSamples PROC NEAR
	push eax
	push ebx
	; --------------------------------------------
	cmp _samplesPerTickHi,0
	je WriteSilence
	; --------------------------------------------	
	push edi
	xor edi,edi ; mix position
	mov byte ptr DidMixAVoice,0
SampleLoop:
	test ebx,ebx ; samples left?
	jle MixDone
	; --------------------------------------------
	; Check if it's time to tick the replayer...
	; --------------------------------------------
	mov edx,_sampleCounterHi
	test edx,edx
	jnz NoNewTick
	cmp _Replayer_SongPlaying,1
	jne NoPlay
	call _Replayer_Tick
NoPlay:	mov edx,_samplesPerTickHi
	mov eax,_samplesPerTickLo
	add _sampleCounterLo,eax
	adc edx,0
	mov _sampleCounterHi,edx
NoNewTick:
	mov ecx,ebx
	cmp ecx,edx
	jle Skip ; samplesLeft > samplesPerTick?
	mov ecx,edx
Skip:	; --------------------------------------------
	; Mix samples (ecx = samples to mix)
	; --------------------------------------------
	call _Mixer_MixVoices
	or DidMixAVoice,al
	add edi,ecx
	add edi,ecx
	; --------------------------------------------
	sub _sampleCounterHi,ecx
	sub ebx,ecx
	jmp SampleLoop
MixDone:
	pop edi		; edi = DMA buffer pointer
	pop ecx		; ecx = number of samples
	shr ecx,4	; number of samples is a multiple of 16, this is safe
	pop eax		; eax = DMA buffer position
	; --------------------------------------------
	; Mix done
	; --------------------------------------------

	cmp byte ptr DidMixAVoice,0
	je WriteSilence	; we didn't mix any voices (all inactive)
	
	; --------------------------------------------
	; Handle Amiga 500 filter (if enabled)
	; --------------------------------------------
	cmp _audioAmiga500,1
	jne NoFilter
	; --------------------------------------------
	push eax
	push ecx
	push edi	
	
	; --------------------------------------------
	; Left channel (or mono)
	; --------------------------------------------
	xor eax,eax
	xor ebx,ebx
	mov ax,_lastFilterSampleL
	mov edi,_lpFilterTab
	mov esi,_mixBufferL
FMLoopL:
	FILTER_M 0
	FILTER_M 1
	FILTER_M 2
	FILTER_M 3
	FILTER_M 4
	FILTER_M 5
	FILTER_M 6
	FILTER_M 7
	FILTER_M 8
	FILTER_M 9
	FILTER_M 10
	FILTER_M 11
	FILTER_M 12
	FILTER_M 13
	FILTER_M 14
	FILTER_M 15
	add esi,2*16
	dec ecx
	jnz FMLoopL
	mov _lastFilterSampleL,ax
	mov ecx,[esp+4]
	
	; --------------------------------------------
	; Right channel (if in stereo mode)
	; --------------------------------------------
	cmp _stereoAudio,1
	jne NoFilterR
	; --------------------------------------------
	mov ax,_lastFilterSampleR
	mov esi,_mixBufferR
FMLoopR:
	FILTER_M 0
	FILTER_M 1
	FILTER_M 2
	FILTER_M 3
	FILTER_M 4
	FILTER_M 5
	FILTER_M 6
	FILTER_M 7
	FILTER_M 8
	FILTER_M 9
	FILTER_M 10
	FILTER_M 11
	FILTER_M 12
	FILTER_M 13
	FILTER_M 14
	FILTER_M 15
	add esi,2*16
	dec ecx
	jnz FMLoopR
	mov _lastFilterSampleR,ax
	; --------------------------------------------
NoFilterR:
	pop edi
	pop ecx
	pop eax
NoFilter:

	; --------------------------------------------
	; Transfer mixed samples to DMA buffer
	; --------------------------------------------
	mov OldStackPointer,esp
	cmp _audio16Bit,1
	jne PostMix8Bit
	; --------------------------------------------
	cmp _stereoSeparation,0
	je PostMix16BitCentered

	; --------------------------------------------
	; Signed 16-bit -> signed 16-bit (stereo)
	; --------------------------------------------
PostMix16BitStereo:
	add ecx,ecx
	mov LoopCounter,cx
	mov ebp,eax		; ebp = DMA position
	mov ecx,_mixBufferL
	mov DMABufPtr,edi
	mov esi,_mixBufferR
	mov esp,_audioSideFactorMul
	; ---------------------
p16LS:	mov edi,DMABufPtr
	add edi,ebp
	DMA16S_M 0
	DMA16S_M 1
	DMA16S_M 2
	DMA16S_M 3
	DMA16S_M 4
	DMA16S_M 5
	DMA16S_M 6
	DMA16S_M 7
	add ecx,8*2
	add esi,8*2
	add ebp,16*2
	and ebp,DMA_BUF_MASK
	dec word ptr LoopCounter
	jnz p16LS
	; ---------------------
	mov esp,OldStackPointer
	retn
	
	; --------------------------------------------
	; Signed 16-bit -> signed 16-bit (stereo, centered)
	; --------------------------------------------	
PostMix16BitCentered:
	xor ebx,ebx
	add ecx,ecx
	mov esp,DMA_BUF_MASK
	mov ebp,ecx
	mov ecx,edi
	mov esi,_mixBufferL
	; ---------------------
p16LM:	lea edi,[ecx+eax]
	DMA16M_M 0
	DMA16M_M 1
	DMA16M_M 2
	DMA16M_M 3
	DMA16M_M 4
	DMA16M_M 5
	DMA16M_M 6
	DMA16M_M 7
	add esi,8*2
	add eax,16*2
	and eax,esp
	dec ebp
	jnz p16LM
	; ---------------------
	mov esp,OldStackPointer
	retn

	; --------------------------------------------
	; Signed 16-bit -> unsigned 8-bit (always mono)
	; --------------------------------------------
PostMix8Bit:
	xor esp,esp
	mov edx,80808080h ; XOR sign flip (4 samples)
	mov ebp,ecx
	mov ecx,edi
	mov esi,_mixBufferL
	; ---------------------
p8L:	lea edi,[ecx+eax]
	DMA8M_M 0
	DMA8M_M 1
	DMA8M_M 2
	DMA8M_M 3
	add esi,16*2
	add eax,16
	and eax,DMA_BUF_MASK
	dec ebp
	jnz p8L
	; ---------------------
	mov esp,OldStackPointer
	retn
	
	; --------------------------------------------
	; Write silence to DMA buffer
	; -------------------------------------------
WriteSilence:
	mov ebp,DMA_BUF_MASK
	cmp _stereoAudio,1
	jne WS0
	add ecx,ecx
WS0:	cmp _audio16Bit,1
	jne Clear8Bit

	; --------------------------------------------
	; Signed 16-bit DMA clear
	; --------------------------------------------
Clear16Bit:
	mov esi,16*2
	xor ebx,ebx
c16L:	lea edx,[edi+eax]
	mov [edx+0],ebx
	mov [edx+4],ebx
	mov [edx+8],ebx
	mov [edx+12],ebx
	mov [edx+16],ebx
	mov [edx+20],ebx
	mov [edx+24],ebx
	mov [edx+28],ebx
	add eax,esi
	and eax,ebp
	dec ecx
	jnz c16L
	retn


	; --------------------------------------------
	; Unsigned 8-bit DMA clear
	; --------------------------------------------
Clear8Bit:
	mov esi,16
	mov ebx,80808080h
c8L:	lea edx,[edi+eax]
	mov [edx+0],ebx
	mov [edx+4],ebx
	mov [edx+8],ebx
	mov [edx+12],ebx
	add eax,esi
	and eax,ebp
	dec ecx
	jnz c8L
	retn

_Audio_MixSamples ENDP

	ALIGN 4
OldStackPointer	dd 0
DMABufPtr	dd 0
LoopCounter	dw 0
DidMixAVoice	db 0

END
