.386p
.MODEL FLAT

; NOTE TO SELF:
;   If using base-EBP to index memory, remember DS segment override (or else
;    you get a crash under Windows since this runs in an interrupt)

	INCLUDE "apaula.i"

	; paula.c
	EXTERN _paulaVolTable:DWORD
	EXTERN _paulaPeriod2Hz_Lo:DWORD
	EXTERN _paulaPeriod2Hz_Hi:DWORD

	; audio.c
	EXTERN _audioOpen:BYTE

	PUBLIC _Paula_ClearVoices
	PUBLIC _Paula_SetDMACON
	PUBLIC _Paula_SetDat
	PUBLIC _Paula_SetLen
	PUBLIC _Paula_SetPer
	PUBLIC _Paula_SetVol

.CODE

	ALIGN 4
_Paula_ClearVoices PROC NEAR ; must not be called from an interrupt!
	push eax
	push ebx
	push ecx
	push esi
	; ---------------------------------------
	mov esi,_paulaOffset
	mov ecx,PAULA_VOICES
	; ---------------------------------------
LL0:	xor eax,eax
	mov v_dmaActive[esi],al
	mov v_lastDelta[esi],eax
	
	dec eax	; eax = -1
	mov v_lastPeriod[esi],eax	; clear period cache	

	mov eax,_paulaNULLSectionOffset
	mov v_newDataPtr[esi],eax
	mov v_dataPtr[esi],eax

	mov eax,2
	mov v_newLength[esi],eax
	mov v_length[esi],eax

	xor eax,eax
	call _Paula_SetPer ; period 0 (-> period 65536)
	call _Paula_SetVol ; volume 0

	add esi,v_sizeof
	dec ecx
	jnz LL0
	; ---------------------------------------
	pop esi
	pop ecx
	pop ebx
	pop eax
	retn
_Paula_ClearVoices ENDP

	ALIGN 4
	; Input:
	;  EAX = DMACON value (word)
_Paula_SetDMACON PROC NEAR
	push eax
	push ebx
	push ecx
	push esi
	mov esi,_paulaOffset
	; ---------------------------------------
	test ax,8000h
	jnz TurnOn
	; ---------------------------------------

TurnOff:
	test al,1
	jz NoCh1Off
	mov byte ptr v_dmaActive[esi],0
NoCh1Off:
	add esi,v_sizeof
	test al,2
	jz NoCh2Off
	mov byte ptr v_dmaActive[esi],0
NoCh2Off:
	add esi,v_sizeof
	test al,4
	jz NoCh3Off
	mov byte ptr v_dmaActive[esi],0
NoCh3Off:
	add esi,v_sizeof
	test al,8
	jz SDCEnd
	mov byte ptr v_dmaActive[esi],0
SDCEnd:	pop esi
	pop ecx
	pop ebx
	pop eax
	retn
	; ---------------------------------------	

TurnOn:
	test al,1
	jz NoCh1On
	call StartDMA
NoCh1On:
	add esi,v_sizeof
	test al,2
	jz NoCh2On
	call StartDMA
NoCh2On:
	add esi,v_sizeof
	test al,4
	jz NoCh3On
	call StartDMA
NoCh3On:
	add esi,v_sizeof
	test al,8
	jz SDCEnd
	call StartDMA
	jmp SDCEnd
	; ---------------------------------------

StartDMA:
	mov ebx,v_newDataPtr[esi]
	test ebx,ebx
	jnz L6
	mov ebx,_paulaNULLSectionOffset
	mov v_newDataPtr[esi],ebx
L6:	mov v_dataPtr[esi],ebx
	; -----------------------------------
	mov ecx,v_newLength[esi]
	cmp ecx,2
	jnb L7
	mov cl,2
	mov v_newLength[esi],ecx
L7:	; -----------------------------------
	mov v_length[esi],ecx
	add ebx,ecx
	mov v_dataEnd[esi],ebx
	xor ebx,ebx
	mov v_frac[esi],ebx
	mov byte ptr v_dmaActive[esi],1
	retn
	
_Paula_SetDMACON ENDP

	; Input:
	;  ESI = Pointer to Paula struct
	;  EAX = Data pointer
	ALIGN 4
_Paula_SetDat PROC NEAR
	push eax
	test eax,eax
	jnz L0
	mov eax,_paulaNULLSectionOffset
L0:	mov v_newDataPtr[esi],eax
	pop eax
	retn
_Paula_SetDat ENDP

	; Input:
	;  ESI = Pointer to Paula struct
	;   AX = Length (in unit of words, 0..65535)
	ALIGN 4
_Paula_SetLen PROC NEAR
	push eax
	movzx eax,ax
	test eax,eax
	jnz L1
	inc eax ; on real Amiga, len0=65536, but we set it to 1 to be safe!
L1:	add eax,eax ; convert from unit of bytes
	mov v_newLength[esi],eax
	pop eax
	retn
_Paula_SetLen ENDP

	; Input:
	;  ESI = Pointer to Paula struct
	;   AX = Period (0..65535)
	ALIGN 4
_Paula_SetPer PROC NEAR
	cmp byte ptr _audioOpen,0
	je PSPEnd	; no audio = no period2Delta var set
	; -----------------------------------
	push eax
	movzx eax,ax
	cmp eax,v_lastPeriod[esi]
	je L4 ; this period has been calculated
	mov v_lastPeriod[esi],eax
	; -----------------------------------
	push ebx
	push ecx
	push edx
	test eax,eax
	jnz L2		; period 0 = period 65536
	mov eax,65536
L2:	cmp eax,113
	jnb L3		; period 113 is the absolute minimum
	mov al,113
L3:	mov ebx,eax
	mov ecx,ebx	; get rounding bias
	shr ecx,1
	mov edx,_paulaPeriod2Hz_Hi
	mov eax,_paulaPeriod2Hz_Lo
	add eax,ecx	; add rounding bias
	adc edx,0
	div ebx		; eax = rounded .31fp delta
	; -----------------------------------
	mov v_lastDelta[esi],eax
	pop edx
	pop ecx
	pop ebx
	; -----------------------------------
L4:	mov eax,v_lastDelta[esi]
	mov v_delta[esi],eax
	pop eax
PSPEnd:	retn
_Paula_SetPer ENDP

	; Input:
	;  ESI = Pointer to Paula struct
	;   AX = Volume (0..64)
	ALIGN 4
_Paula_SetVol PROC NEAR
	cmp byte ptr _audioOpen,0	
	je PSVEnd	; no audio = no vol LUT
	push eax
	and eax,127
	cmp al,64
	jna L5
	mov al,64
L5:	mov v_volume[esi],al
	shl eax,8+1
	add eax,_paulaVolTable
	shr eax,1
	mov v_volTabPtr[esi],eax
	pop eax
PSVEnd:	retn
_Paula_SetVol ENDP

END
