#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include "paula.h"
#include "audio.h"
#include "mixer.h"
#include "replayer.h"

static int16_t *volTablePtr;

// globals
int8_t paulaNULLSection[64]; // must be at least 2+SCOPE_W long
paula_t paula[PAULA_VOICES];
int16_t *paulaVolTable;
uint32_t paulaNULLSectionOffset, paulaOffset, paulaPeriod2Hz_Hi, paulaPeriod2Hz_Lo;

bool Paula_Init(void)
{
	memset(paula, 0, sizeof (paula));
	memset(paulaNULLSection, 0, sizeof (paulaNULLSection));

	paulaOffset = (uint32_t)paula;
	paulaNULLSectionOffset = (uint32_t)paulaNULLSection;

	if (audioOpen)
	{
		// calculate 64-bit dividend for calculating period to delta
		double dPeriod2Hz = ((double)PAULA_PAL_CLK / audio.dOutputFreq) * (UINT32_MAX+1.0);
		uint64_t period2Hz64 = (uint64_t)(dPeriod2Hz + 0.5);
		paulaPeriod2Hz_Hi = period2Hz64 >> 32;
		paulaPeriod2Hz_Lo = period2Hz64 & 0xFFFFFFFF;
		// ----------------------------------------------------------

		// allocate memory for volume table (address must be a multiple of 512 for speed reasons)
		volTablePtr = (int16_t *)malloc(((PAULA_MAX_VOL+1) * 256 * sizeof (int16_t)) + 512);
		if (volTablePtr == NULL)
			return false;

		paulaVolTable = (int16_t *)(((uint32_t)volTablePtr + 511) & ~511); // aligned
		Mixer_GenerateVolTable(paulaVolTable);
	}

	Paula_ClearVoices();

	return Replayer_InitBPMTable();
}

void Paula_Free(void)
{
	if (volTablePtr != NULL)
	{
		free(volTablePtr);
		volTablePtr = NULL;
	}

	Replayer_FreeBPMTable();
}
