; Mouse pointer routines for PT4DOS.
;
; NOTE: These are meant to only be used on a dedicated bitplane.
;       In this case, it's bitplane 4.
;

.386p
.MODEL FLAT

	INCLUDE "../hardware/avideo.i"

MOUSE_SPEED	EQU 5 ; 1..16
POINTER_W	EQU 8
POINTER_H	EQU 13

; mouse_t struct
m_x		EQU 0
m_y		EQU 4
m_leftbtn	EQU 8
m_rightbtn	EQU 9
m_clickhandled	EQU 10

	EXTERN __GETDS:PROC

	; main.c
	EXTERN _trackerRunning:BYTE
	EXTERN _mouseStructPtr:DWORD

	PUBLIC _Mouse_Callback
	PUBLIC _Mouse_Update
	
.CODE

	; Mouse callback (when moving mouse or pressing/releasing buttons)
	;
	; NOTES:
	;  - This is far callback called from within the main mouse interrupt.
	;    Thus, we exit this routine with RETF (NOT with IRET).
	;  - No need to preserve registers in this callback.
	;  - DS must be set (points to internal driver data segment on enter)
	;
	ALIGN 4
_Mouse_Callback PROC FAR
	call __GETDS
	; -----------------------------
	cmp _trackerRunning,0
	je miEnd
	; -----------------------------
	mov al,1
	xchg al,updateOngoing
	test al,al
	jnz _MUEnd ; a Mouse_Update() call is already ongoing from the program main thread
	; -----------------------------
   	; set mouse position and button states
   	; -----------------------------
   	mov esi,_mouseStructPtr
	bt bx,0
	setc m_leftbtn[esi]
	bt bx,1
	setc m_rightbtn[esi]
	; -----------------------------
	mov ax,000Bh ; read x/y mickeys (si/di in callback is broken..?)
	int 33h
	; -----------------------------
	movsx ebx,cx	
	imul ebx,MOUSE_SPEED
	mov eax,m_x[esi]
	shl eax,4
	add eax,mx_frac
	add ebx,eax
	mov eax,ebx
	and eax,15
	mov mx_frac,eax
	sar ebx,4
	; ebx = mouse X
	; -----------------------------
	movsx ecx,dx
	imul ecx,MOUSE_SPEED		
	mov eax,m_y[esi]
	shl eax,4
	add eax,my_frac
	add ecx,eax
	mov eax,ecx
	and eax,15
	mov my_frac,eax
	sar ecx,4
	; ecx = mouse Y
	; -----------------------------
	
	; -----------------------------
	; clamp mouse position to screen space
	; -----------------------------
	mov eax,SCREEN_W-1
	mov edx,SCREEN_H-1
	test ebx,ebx
	jge _XOK1
	xor ebx,ebx
_XOK1:	cmp ebx,eax
	jle _XOK2
	mov ebx,eax
_XOK2:	test ecx,ecx
	jge _YOK1
	xor ecx,ecx
_YOK1:	cmp ecx,edx
	jle _YOK2
	mov ecx,edx
_YOK2:	; -----------------------------
	mov m_x[esi],ebx
	mov m_y[esi],ecx
	; -----------------------------
	mov ebp,oldX
	mov edi,oldY
	mov oldX,ebx	; update oldX
	mov oldY,ecx	; update oldY
	; -----------------------------
	cmp ebx,ebp	; current X == old X?
	jne _DrawMouse
	cmp ecx,edi	; current Y == old Y?
	je _NoDraw
_DrawMouse:
	; -----------------------------
	; backup VGA registers
	; -----------------------------
	mov dx,3C4h
	mov ax,0202h
	out dx,al
	inc dx
	in al,dx
	xchg al,ah
	mov oldPlaneMask,ax
	; -----------------------------
	mov dl,0CEh
	mov ax,0505h
	out dx,al
	inc dx
	in al,dx
	dec dx
	xchg al,ah
	mov oldWriteMode,ax
	; -----------------------------
	mov ax,0808h
	out dx,al
	inc dx
	in al,dx
	dec dx
	xchg al,ah
	mov oldBitMask,ax	
	; -----------------------------
	mov ax,0303h
	out dx,al
	inc dx
	in al,dx
	dec dx
	xchg al,ah
	mov oldLatchMode,ax
	; -----------------------------
	; set new VGA registers
	; -----------------------------
	mov dx,3C4h
	mov ax,0802h
	out dx,ax
	; -----------------------------
	mov dl,0CEh
	mov ax,0005h
	out dx,ax
	; -----------------------------
	mov al,03h
	out dx,ax
	; -----------------------------
	mov ax,0FF08h
	out dx,ax
	; -----------------------------
	; erase last mouse pointer (ebp = old X, edi = old Y)
	; -----------------------------
	cmp byte ptr wasDrawn,0
	je _NoErase
	; -------------------------------
	; setup frame buffer pointer
	; -------------------------------
	shl edi,3
	lea edi,[edi+edi*4]			; edx *= 40 (SCREEN_W/8)
	shr ebp,3 				; eax /= 8
	lea edi,[VGA_FRAMEBUFFER+edi+ebp]
	mov edx,(SCREEN_W/8)*4
	xor eax,eax
	; -------------------------------
	mov [edi+((SCREEN_W/8)*0)],ax
	mov [edi+((SCREEN_W/8)*1)],ax
	mov [edi+((SCREEN_W/8)*2)],ax
	mov [edi+((SCREEN_W/8)*3)],ax
	add edi,edx
	mov [edi+((SCREEN_W/8)*0)],ax
	mov [edi+((SCREEN_W/8)*1)],ax
	mov [edi+((SCREEN_W/8)*2)],ax
	mov [edi+((SCREEN_W/8)*3)],ax
	add edi,edx
	mov [edi+((SCREEN_W/8)*0)],ax
	mov [edi+((SCREEN_W/8)*1)],ax
	mov [edi+((SCREEN_W/8)*2)],ax
	mov [edi+((SCREEN_W/8)*3)],ax
	add edi,edx
	mov [edi+((SCREEN_W/8)*0)],ax
_NoErase:; -----------------------------
	; draw mouse pointer (ebx = x, ecx = y)
	; -----------------------------
	
	; -------------------------------
	; setup bitplane data pointer
	; -------------------------------
	mov edx,ebx
	and edx,8-1				; edx = mx & 7
	shl edx,5
	lea edx,[PointerData+edx]
	cmp ebx,SCREEN_W-8
	jb _NonClipped
	add edx,16*2*8				; used masked pointer data
_NonClipped:
	; -------------------------------
	; setup frame buffer pointer
	; -------------------------------
	shl ecx,3
	lea ecx,[ecx+ecx*4]			; ecx *= 40 (SCREEN_W/8)
	shr ebx,3 				; ebx = x / 8
	lea edi,[VGA_FRAMEBUFFER+ecx+ebx]
	mov ebp,(SCREEN_W/8)*4
	; -------------------------------
	mov eax,[edx]
	mov [edi+((SCREEN_W/8)*0)],ax
	shr eax,16
	mov [edi+((SCREEN_W/8)*1)],ax
	mov eax,[edx+4]
	mov [edi+((SCREEN_W/8)*2)],ax
	shr eax,16
	mov [edi+((SCREEN_W/8)*3)],ax
	; -------------------------------
	add edi,ebp
	; -------------------------------
	mov eax,[edx+8]
	mov [edi+((SCREEN_W/8)*0)],ax
	shr eax,16
	mov [edi+((SCREEN_W/8)*1)],ax
	mov eax,[edx+12]
	mov [edi+((SCREEN_W/8)*2)],ax
	shr eax,16
	mov [edi+((SCREEN_W/8)*3)],ax
	; -------------------------------
	add edi,ebp
	; -------------------------------
	mov eax,[edx+16]
	mov [edi+((SCREEN_W/8)*0)],ax
	shr eax,16
	mov [edi+((SCREEN_W/8)*1)],ax
	mov eax,[edx+20]
	mov [edi+((SCREEN_W/8)*2)],ax
	shr eax,16
	mov [edi+((SCREEN_W/8)*3)],ax
	; -------------------------------
	add edi,ebp
	; -------------------------------
	mov ax,[edx+24]
	mov [edi+((SCREEN_W/8)*0)],ax
	; -------------------------------
	mov byte ptr wasDrawn,1	
	; -----------------------------
	; restore VGA registers
	; -----------------------------
	mov dx,3C4h
	mov ax,oldPlaneMask
	out dx,ax
	; -----------------------------
	mov dl,0CEh
	mov ax,oldWriteMode
	out dx,ax
	; -----------------------------
	mov ax,oldBitMask
	out dx,ax
	; -----------------------------
	mov ax,oldLatchMode
	out dx,ax
_NoDraw:
	; -----------------------------
	cmp byte ptr m_leftbtn[esi],1
	je _MSkip
	cmp byte ptr m_rightbtn[esi],1
	je _MSkip
	mov byte ptr m_clickhandled[esi],0
	; -----------------------------
_MSkip:	mov byte ptr updateOngoing,0
_MUEnd:	; -----------------------------
miEnd:	retf
_Mouse_Callback ENDP

	; This routine is ONLY called from the program main thread
	;
	ALIGN 4
_Mouse_Update PROC NEAR
	cmp _trackerRunning,0
	je MUX
	; -----------------------------
	pushad
	mov al,1
	xchg al,updateOngoing
	test al,al
	jnz MUEnd ; this routine is already being called!
   	; -----------------------------
   	; set mouse button states
   	; -----------------------------
   	mov esi,_mouseStructPtr
	; -----------------------------
	mov ax,000Bh ; read x/y mickeys
	int 33h
	; -----------------------------
	movsx ebx,cx	
	imul ebx,MOUSE_SPEED
	mov eax,m_x[esi]
	shl eax,4
	add eax,mx_frac
	add ebx,eax
	mov eax,ebx
	and eax,15
	mov mx_frac,eax
	sar ebx,4
	; ebx = mouse X
	; -----------------------------
	movsx ecx,dx
	imul ecx,MOUSE_SPEED		
	mov eax,m_y[esi]
	shl eax,4
	add eax,my_frac
	add ecx,eax
	mov eax,ecx
	and eax,15
	mov my_frac,eax
	sar ecx,4
	; ecx = mouse Y
	; -----------------------------
	
	; -----------------------------
	; clamp mouse position to screen space
	; -----------------------------
	mov eax,SCREEN_W-1
	mov edx,SCREEN_H-1
	test ebx,ebx
	jge XOK1
	xor ebx,ebx
XOK1:	cmp ebx,eax
	jle XOK2
	mov ebx,eax
XOK2:	test ecx,ecx
	jge YOK1
	xor ecx,ecx
YOK1:	cmp ecx,edx
	jle YOK2
	mov ecx,edx
YOK2:	; -----------------------------
	mov m_x[esi],ebx
	mov m_y[esi],ecx
	; -----------------------------
	mov ebp,oldX
	mov edi,oldY
	mov oldX,ebx	; update oldX
	mov oldY,ecx	; update oldY
	; -----------------------------
	cmp ebx,ebp	; current X == old X?
	jne DrawMouse
	cmp ecx,edi	; current Y == old Y?
	je NoDraw
DrawMouse:
	; -----------------------------
	; set write bitplane
	; -----------------------------
	mov dx,3C4h
	mov ax,0802h
	out dx,ax
	; -----------------------------
	; erase last mouse pointer (ebp = old X, edi = old Y)
	; -----------------------------
	cmp byte ptr wasDrawn,0
	je NoErase
	; -------------------------------
	; setup frame buffer pointer
	; -------------------------------
	shl edi,3
	lea edi,[edi+edi*4]			; edx *= 40 (SCREEN_W/8)
	shr ebp,3 				; eax /= 8
	lea edi,[VGA_FRAMEBUFFER+edi+ebp]
	mov edx,(SCREEN_W/8)*4
	xor eax,eax
	; -------------------------------
	mov [edi+((SCREEN_W/8)*0)],ax
	mov [edi+((SCREEN_W/8)*1)],ax
	mov [edi+((SCREEN_W/8)*2)],ax
	mov [edi+((SCREEN_W/8)*3)],ax
	add edi,edx
	mov [edi+((SCREEN_W/8)*0)],ax
	mov [edi+((SCREEN_W/8)*1)],ax
	mov [edi+((SCREEN_W/8)*2)],ax
	mov [edi+((SCREEN_W/8)*3)],ax
	add edi,edx
	mov [edi+((SCREEN_W/8)*0)],ax
	mov [edi+((SCREEN_W/8)*1)],ax
	mov [edi+((SCREEN_W/8)*2)],ax
	mov [edi+((SCREEN_W/8)*3)],ax
	add edi,edx
	mov [edi+((SCREEN_W/8)*0)],ax
NoErase:; -----------------------------
	; draw mouse pointer (ebx = x, ecx = y)
	; -----------------------------
	
	; -------------------------------
	; setup bitplane data pointer
	; -------------------------------
	mov edx,ebx
	and edx,8-1				; edx = mx & 7
	shl edx,5
	lea edx,[PointerData+edx]
	cmp ebx,SCREEN_W-8
	jb NonClipped
	add edx,16*2*8				; used masked pointer data
NonClipped:
	; -------------------------------
	; setup frame buffer pointer
	; -------------------------------
	shl ecx,3
	lea ecx,[ecx+ecx*4]			; ecx *= 40 (SCREEN_W/8)
	shr ebx,3 				; ebx = x / 8
	lea edi,[VGA_FRAMEBUFFER+ecx+ebx]
	mov ebp,(SCREEN_W/8)*4
	; -------------------------------
	mov eax,[edx]
	mov [edi+((SCREEN_W/8)*0)],ax
	shr eax,16
	mov [edi+((SCREEN_W/8)*1)],ax
	mov eax,[edx+4]
	mov [edi+((SCREEN_W/8)*2)],ax
	shr eax,16
	mov [edi+((SCREEN_W/8)*3)],ax
	; -------------------------------
	add edi,ebp
	; -------------------------------
	mov eax,[edx+8]
	mov [edi+((SCREEN_W/8)*0)],ax
	shr eax,16
	mov [edi+((SCREEN_W/8)*1)],ax
	mov eax,[edx+12]
	mov [edi+((SCREEN_W/8)*2)],ax
	shr eax,16
	mov [edi+((SCREEN_W/8)*3)],ax
	; -------------------------------
	add edi,ebp
	; -------------------------------
	mov eax,[edx+16]
	mov [edi+((SCREEN_W/8)*0)],ax
	shr eax,16
	mov [edi+((SCREEN_W/8)*1)],ax
	mov eax,[edx+20]
	mov [edi+((SCREEN_W/8)*2)],ax
	shr eax,16
	mov [edi+((SCREEN_W/8)*3)],ax
	; -------------------------------
	add edi,ebp
	; -------------------------------
	mov ax,[edx+24]
	mov [edi+((SCREEN_W/8)*0)],ax
	; -------------------------------
	mov byte ptr wasDrawn,1	
NoDraw: ; -----------------------------
	; read mouse button states
	; -----------------------------
	mov ax,3
	int 33h
	bt bx,0
	setc m_leftbtn[esi]
	bt bx,1
	setc m_rightbtn[esi]
	; -----------------------------
	cmp byte ptr m_leftbtn[esi],1
	je MSkip
	cmp byte ptr m_rightbtn[esi],1
	je MSkip
	mov byte ptr m_clickhandled[esi],0
	; -----------------------------
MSkip:	mov byte ptr updateOngoing,0
MUEnd:	popad
MUX:	retn
_Mouse_Update ENDP

	ALIGN 4
PointerData:
	; x & 7 == 0
	dw 00080h,000C0h,000E0h,000F0h,000F8h,000FCh,000FEh,000FFh
	dw 000F8h,000D8h,0009Ch,0000Ch,0000Ch,00000h,00000h,00000h

	; x & 7 == 1
	dw 00040h,00060h,00070h,00078h,0007Ch,0007Eh,0007Fh,0807Fh
	dw 0007Ch,0006Ch,0004Eh,00006h,00006h,00000h,00000h,00000h

	; x & 7 == 2
	dw 00020h,00030h,00038h,0003Ch,0003Eh,0003Fh,0803Fh,0C03Fh
	dw 0003Eh,00036h,00027h,00003h,00003h,00000h,00000h,00000h

	; x & 7 == 3
	dw 00010h,00018h,0001Ch,0001Eh,0001Fh,0801Fh,0C01Fh,0E01Fh
	dw 0001Fh,0001Bh,08013h,08001h,08001h,00000h,00000h,00000h

	; x & 7 == 4
	dw 00008h,0000Ch,0000Eh,0000Fh,0800Fh,0C00Fh,0E00Fh,0F00Fh
	dw 0800Fh,0800Dh,0C009h,0C000h,0C000h,00000h,00000h,00000h

	; x & 7 == 5
	dw 00004h,00006h,00007h,08007h,0C007h,0E007h,0F007h,0F807h
	dw 0C007h,0C006h,0E004h,06000h,06000h,00000h,00000h,00000h

	; x & 7 == 6
	dw 00002h,00003h,08003h,0C003h,0E003h,0F003h,0F803h,0FC03h
	dw 0E003h,06003h,07002h,03000h,03000h,00000h,00000h,00000h

	; x & 7 == 7
	dw 00001h,08001h,0C001h,0E001h,0F001h,0F801h,0FC01h,0FE01h
	dw 0F001h,0B001h,03801h,01800h,01800h,00000h,00000h,00000h

	; x & 7 == 0 (outside of right edge)
	dw 00080h,000C0h,000E0h,000F0h,000F8h,000FCh,000FEh,000FFh
	dw 000F8h,000D8h,0009Ch,0000Ch,0000Ch,00000h,00000h,00000h

	; x & 7 == 1 (outside of right edge)
	dw 00040h,00060h,00070h,00078h,0007Ch,0007Eh,0007Fh,0007Fh
	dw 0007Ch,0006Ch,0004Eh,00006h,00006h,00000h,00000h,00000h

	; x & 7 == 2 (outside of right edge)
	dw 00020h,00030h,00038h,0003Ch,0003Eh,0003Fh,0003Fh,0003Fh
	dw 0003Eh,00036h,00027h,00003h,00003h,00000h,00000h,00000h

	; x & 7 == 3 (outside of right edge)
	dw 00010h,00018h,0001Ch,0001Eh,0001Fh,0001Fh,0001Fh,0001Fh
	dw 0001Fh,0001Bh,00013h,00001h,00001h,00000h,00000h,00000h

	; x & 7 == 4 (outside of right edge)
	dw 00008h,0000Ch,0000Eh,0000Fh,0000Fh,0000Fh,0000Fh,0000Fh
	dw 0000Fh,0000Dh,00009h,00000h,00000h,00000h,00000h,00000h

	; x & 7 == 5 (outside of right edge)
	dw 00004h,00006h,00007h,00007h,00007h,00007h,00007h,00007h
	dw 00007h,00006h,00004h,00000h,00000h,00000h,00000h,00000h

	; x & 7 == 6 (outside of right edge)
	dw 00002h,00003h,00003h,00003h,00003h,00003h,00003h,00003h
	dw 00003h,00003h,00002h,00000h,00000h,00000h,00000h,00000h

	; x & 7 == 7 (outside of right edge)
	dw 00001h,00001h,00001h,00001h,00001h,00001h,00001h,00001h
	dw 00001h,00001h,00001h,00000h,00000h,00000h,00000h,00000h

	ALIGN 4
mx_frac		dd 0
my_frac		dd 0
oldX		dd -1
oldY		dd -1
oldPlaneMask	dw 0
oldWriteMode	dw 0
oldBitMask	dw 0
oldLatchMode	dw 0
updateOngoing	db 0
wasDrawn	db 0

END
