#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <conio.h>
#include <dos.h>
#include "irq.h"
#include "keyb.h"
#include "mouse.h"
#include "video.h"
#include "../ui.h" // UI_HandleKeyPress()

#define KEYB_IRQ 1
#define KEYB_BUF_SIZE 32

keyb_t keyb;

static bool testNextKey = false;
static uint8_t keyBuf[KEYB_BUF_SIZE], writePtr, readPtr;

static void __interrupt int9Vector(void)
{
	const uint8_t keyByte = inp(0x60);
	uint8_t scancode = keyByte & 127;
	
	//if (scancode == 0)
	//	return;

	if (keyByte == 98 || keyByte == 224) // for right Alt and right Ctrl
	{
		testNextKey = true;
	}
	else
	{
		if (testNextKey)
		{
			if (scancode == KEY_CTRL)
				scancode = KEY_RCTRL;
			else if (scancode == KEY_ALT)
				scancode = KEY_RALT;

			testNextKey = false;
		}

		if (keyByte & 128) // key released?
		{
			if (scancode == KEY_CTRL)
				keyb.ctrlPressed = false;
			else if (scancode == KEY_ALT)
				keyb.altPressed  = false;
			else if (scancode == KEY_LSHIFT)
				keyb.shiftPressed = false;
		}
		else
		{
			uint8_t curKeyPtr = writePtr;

			writePtr = (writePtr + 1) & (KEYB_BUF_SIZE-1);
			if (writePtr != readPtr)
				keyBuf[curKeyPtr] = scancode;
				
			if (scancode == KEY_CTRL)
				keyb.ctrlPressed = true;
			else if (scancode == KEY_ALT)
				keyb.altPressed  = true;
			else if (scancode == KEY_LSHIFT)
				keyb.shiftPressed = true;
		}
	}

	outp(PIC1MODE, PICEOI); // send EOI to PIC chip #1 (keyb IRQ == 1)
}

void Keyb_EmptyBuffer(void)
{
	readPtr = writePtr = 0;
}

bool Keyb_KeyPressed(void)
{
	return (writePtr != readPtr);
}

uint8_t Keyb_ReadKey(void)
{
	uint8_t scancode;

	if (writePtr == readPtr)
		return 0; // no key in buffer
		
	scancode = keyBuf[readPtr];
	readPtr = (readPtr + 1) & (KEYB_BUF_SIZE-1);

	return scancode;
}

bool Keyb_Initialize(void)
{
	if (IRQ_SetIntVector(KEYB_IRQ, int9Vector))
	{
		Keyb_EmptyBuffer();
		IRQ_Enable(KEYB_IRQ);
		return true;
	}

	return false;
}

bool Keyb_Close(void)
{
	bool result = IRQ_RestoreIntVector(KEYB_IRQ);

	// clear potentially stuck BIOS key modifier states (this happens)
	*(uint16_t *)0x417 = (*(uint16_t *)0x417) & 0xFCF0;

	return result;
}

void Keyb_HandleInput(void)
{
	if (Keyb_KeyPressed())
	{
		uint8_t scancode = Keyb_ReadKey();
		if (!mouse.blockInput) // UI is not busy?
			UI_HandleKeyPress(scancode);
	}
}
