#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <dos.h>
#include <conio.h>
#include <i86.h>
#include "../ui.h"
#include "../audio/replayer.h"
#include "video.h"
#include "mouse.h"

#define MOUSE_DRIVER_INSTALLED 0xFFFF

static bool mouseOpen;
static uint8_t blockCounter;
static uint8_t initialDelayCounter, repeatDelayCounter, repeatDelayDuration;

static uint16_t dX1, dX2, dY1, dY2;
static uint16_t oldMickeysX, oldMickeysY, oldDoubleSpeedThreshold;
static void (*repeatCallback)();

mouse_t mouse;

extern void Mouse_Callback(void); // amouse.asm

void Mouse_SetupGUIButtonRepeat(uint16_t x1, uint16_t x2, uint16_t y1, uint16_t y2, void (*callback)(), uint8_t initialDelay, uint8_t delayDuration)
{
	dX1 = x1; dX2 = x2;
	dY1 = y1; dY2 = y2;
	initialDelayCounter = initialDelay;
	repeatDelayDuration = delayDuration;
	repeatCallback = callback;
	repeatDelayCounter = 0;
}

void Mouse_HandleButtonRepeat(void)
{
	if (repeatCallback != NULL && mouse.x >= dX1 && mouse.x < dX2 && mouse.y >= dY1 && mouse.y < dY2)
	{	
		if (initialDelayCounter == 0)
		{
			if (repeatDelayCounter++ >= repeatDelayDuration)
			{
				repeatDelayCounter = 0;
				repeatCallback();
			}
		}
		else
		{
			initialDelayCounter--;
		}
	}
}

void Mouse_SetDiskAccessColor(void)
{
	Video_SetMouseColor(21, 55, 8);
}

void Mouse_UpdateColor(void)
{
	if (mouse.blockInput)
		Video_SetMouseColor(55, 0, 4);
	else if (ui.askBoxShown)
		Video_SetMouseColor(51, 8, 38);
	else if (Replayer_PlayMode == PLAY_MODE_SONG || Replayer_PlayMode == PLAY_MODE_PATTERN)
		Video_SetMouseColor(63, 29, 4);
	else if (Replayer_PlayMode == PLAY_MODE_RECORD || Replayer_PlayMode == PLAY_MODE_EDIT)
		Video_SetMouseColor(0, 46, 63);
	else // idle
		Video_SetMouseColor(63, 63, 63);
}

void Mouse_Error(void)
{
	mouse.blockInput = true;
	Mouse_UpdateColor();
	blockCounter = ERR_DURATION;
}

bool Mouse_Initialize(void)
{
#define INSTRUCTION_IRET 0xCD
	struct SREGS sr;
	union REGS r;
	void (interrupt *intHandler)(void);
	
	// check if interrupt 0x33 is installed before interacting with it
	
	intHandler = _dos_getvect(0x33);
	if (intHandler != NULL && *(uint8_t *)intHandler != INSTRUCTION_IRET)
	{
		// detect (and reset) mouse
		r.w.ax = 0x0000;
		int386(0x33, &r, &r);

		if (r.w.ax == MOUSE_DRIVER_INSTALLED)
		{
			// mouse is present, center internal mouse coordinates
			mouse.x = (SCREEN_W/2) - (POINTER_W/2);
			mouse.y = (SCREEN_H/2) - (POINTER_H/2);
			
			// get mouse settings
			r.w.ax = 0x001B;
			int386(0x33, &r, &r);
			oldMickeysX = r.w.bx;
			oldMickeysY = r.w.cx;				
			oldDoubleSpeedThreshold = r.w.dx;
			
			// set new aspect-ratio corrected settings
			if (oldMickeysX > 0)
			{
				r.w.ax = 0x001A;
				r.w.bx = oldMickeysX;
				r.w.cx = oldMickeysX;
				r.w.dx = oldDoubleSpeedThreshold;
				int386(0x33, &r, &r);
			}
			
			// set mouse callback
			segread(&sr);
			r.w.ax = 0x000C;
			r.w.cx = 0x001F; // event mask (callback on: movement, left/right button)
			r.x.edx = FP_OFF(Mouse_Callback);
			sr.es = FP_SEG(Mouse_Callback);
			int386x(0x33, &r, &r, &sr);

			mouseOpen = true;
			return true;
		}
	}

	printf("ERROR: Couldn't detect an installed mouse driver!\n");
	return false;
}

void Mouse_Close(void)
{
	struct SREGS sr;
	union REGS r;

	if (mouseOpen)
	{
		// disable our mouse callback
		segread(&sr);
		r.w.ax = 0x000C;
		r.w.cx = 0;
		r.x.edx = FP_OFF(Mouse_Callback);
		sr.es = FP_SEG(Mouse_Callback);
		int386x(0x33, &r, &r, &sr);

		// reset mouse
		r.w.ax = 0x0000;
		int386(0x33, &r, &r);
		
		// set back old mouse settings
		if (oldMickeysX > 0 && oldMickeysY > 0)
		{
			r.w.ax = 0x001A;
			r.w.bx = oldMickeysX;
			r.w.cx = oldMickeysY;
			r.w.dx = oldDoubleSpeedThreshold;
			int386(0x33, &r, &r);
		}

		mouseOpen = false;
	}
}

void Mouse_HandleInput(void)
{
	// handle busy mouse (red color for 1 second + input blocking)
	if (mouse.blockInput)
	{
		blockCounter--;
		if (blockCounter == 0)
		{	
			mouse.blockInput = false;
			Mouse_UpdateColor();
		}

		return;
	}

	if (!mouse.leftButtonDown && repeatCallback != NULL)
	{
		repeatCallback = NULL;
		mouse.clickHandled = false;
	}
		
	if (!mouse.clickHandled && mouse.leftButtonDown)
	{
		mouse.clickHandled = true;
		UI_HandleMouseClick();
	}
}
