// some of this code is partly based on Fasttracker II's code

#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <conio.h>
#include "pit.h"
#include "irq.h"
#include "video.h"
#include "../audio/audio.h"

#define TIMER_IRQ 0

// globals
uint32_t dmaMixPos, oldInt8VectorDelta, oldInt8VectorFrac;
uint32_t vblankAudioSamples15; // * 1.5
uint32_t vblankAudioSamples25; // * 2.5
void __interrupt (*oldInt8Vector)(void);

// apitint.asm
extern void __interrupt int8Vector(void);
extern void __interrupt int8VectorNoAudio(void);

void PIT_StartVblankInterrupt(void)
{
	double dOldTimerHz = PIT_CLK / 65536.0; // ~18.2065Hz (clock for original int0 on PIT channel 0)
	double dNewTimerHz = PIT_CLK / (double)vblankPitTicks;

	oldInt8Vector = IRQ_GetIntVector(TIMER_IRQ);
	oldInt8VectorDelta = (uint32_t)(((dOldTimerHz / dNewTimerHz) * (UINT32_MAX+1.0)) + 0.5); // rounded 0.32fp
	oldInt8VectorFrac = 0;

	if (audioOpen)
	{
		// truncate when calculating these three values. do not round!
		const uint32_t vblankAudioSamples = (uint32_t)(audio.dOutputFreq / dNewTimerHz);
		vblankAudioSamples15 = vblankAudioSamples + (vblankAudioSamples >> 1); // *1.5
		vblankAudioSamples25 = (vblankAudioSamples << 1) + (vblankAudioSamples >> 1); // *2.5

		dmaMixPos = (3 * (vblankAudioSamples << audio.sampleShift)) & ~31;
		if (stereoAudio)
			dmaMixPos &= ~63;

		IRQ_SetIntVector(TIMER_IRQ, int8Vector);
	}
	else
	{
		// PT4DOS was started with no audio
		IRQ_SetIntVector(TIMER_IRQ, int8VectorNoAudio);
	}

	// set PIT channel 0 timer
	_disable();
	outp(0x43, 0x36); // channel 0, mode 3
	outp(0x40, vblankPitTicks & 0xFF);
	outp(0x40, vblankPitTicks >> 8);
	Video_WaitForVblank();
	IRQ_Enable(TIMER_IRQ);
	_enable();

	vblankPitTimerEnabled = true;
}

void PIT_StopVblankInterrupt(void)
{
	if (vblankPitTimerEnabled) 
	{
		// restore PIT channel 0 back to default (BIOS timer)
		_disable();
		outp(0x43, 0x36); // channel 0, mode 3
		outp(0x40, 0); // ~18.2065Hz (0 = period 65536)
		outp(0x40, 0);
		_enable();

		IRQ_RestoreIntVector(TIMER_IRQ);
		vblankPitTimerEnabled = false;
	}
}
