#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <i86.h>
#include <graph.h>
#include "video.h"
#include "mouse.h"

#define VGA_PAL_NUM 256

const pal_t origTrackerPal[PAL_NUM_COLORS] =
{
	{  0,  0,  0 }, // PAL_BACKGRD
	{ 46, 46, 46 }, // PAL_BORDER
	{ 34, 34, 34 }, // PAL_GENBKG
	{ 21, 21, 21 }, // PAL_GENBKG2
	{ 63, 55,  0 }, // PAL_QADSCP
	{ 55,  0, 17 }, // PAL_PATCURSOR
	{  0,  0,  0 }, // PAL_GENTXT
	{ 12, 17, 63 }  // PAL_PATTXT
};

static uint8_t origVideoMode, tmpPal[VGA_PAL_NUM*3];

// globalized
volatile bool vblankPitTimerEnabled, updateMousePalette;
uint8_t mouseR, mouseG, mouseB;
uint16_t vblankPitTicks;

bool Video_IsColorVGA(void)
{
	union REGS r;

	r.w.ax = 0x1A00;
	int386(0x10, &r, &r);

	if (r.h.al != 0x1A || ((r.h.bl != 8) && r.h.bl <= 12))
		return false;

	return true;
}

void Video_SetBackgroundColor(uint8_t r, uint8_t g, uint8_t b)
{
__asm
{
	push eax
	push edx
	mov dx,0x3C8
	xor eax,eax
	out dx,al
	inc edx
	mov al,r
	out dx,al
	mov al,g
	out dx,al
	mov al,b
	out dx,al
	pop edx
	pop eax
};
}

void Video_RestoreBackgroundColor(void)
{
__asm
{
	push eax
	push edx
	mov dx,0x3C8
	xor eax,eax
	out dx,al
	inc edx
	out dx,al
	out dx,al
	out dx,al
	pop edx
	pop eax
};
}

void Video_GetVideoMode(void)
{
	union REGS r;

	r.w.ax = 0x0F00;
	int386(0x10, &r, &r);

	origVideoMode = r.h.al;
}

void Video_SetPlaneBitmask(uint8_t bitmask)
{
	outpw(0x3C4, (bitmask << 8) | 0x02);
}

void Video_SelectWritePlane(uint8_t plane)
{
	outpw(0x3C4, ((1 << plane) << 8) | 0x02);
}

void Video_SelectReadPlane(uint8_t plane)
{
	outpw(0x3CE, (plane << 8) | 0x04);
}

void Video_ResetTrackerPalette(void)
{
	int32_t i;

	const pal_t *p = origTrackerPal;
	for (i = 0; i < PAL_NUM_COLORS; i++, p++)
		Video_SetPalColor(i, p->r, p->g, p->b);

	Mouse_UpdateColor();
}

void Video_SetNewVideoMode(void)
{
	union REGS r;

	// enter 320x200 16-color (planar) @ 70Hz
	r.w.ax = 0x000D;
	int386(0x10, &r, &r);

	_disable();

	// tweak current video mode into 320x240 60Hz
	outp(0x3C2, 0xE3);
	outpw(0x3D4, 0x2C11); // turn off write protection
	outpw(0x3D4, 0x0D06); // vertical total
	outpw(0x3D4, 0x3E07); // overflow register
	outpw(0x3D4, 0xEA10); // vertical retrace start
	outpw(0x3D4, 0xAC11); // vertical retrace end & write protection
	outpw(0x3D4, 0xDF12); // vertical display enable end
	outpw(0x3D4, 0xE715); // vertical blanking start
	outpw(0x3D4, 0x0616); // vertical blanking end

	// reset VGA regs (just in case)
	outpw(0x3C4, 0x0102); // select 1st bitplane for writing
	outpw(0x3CE, 0x0004); // select 1st bitplane for reading
	outpw(0x3CE, 0x0005); // write mode #0
	outpw(0x3CE, 0xFF08); // normal bitmask
	outpw(0x3CE, 0x0003); // no data rotate

	_enable();

	Video_ResetTrackerPalette();
}

void Video_ResetVideoMode(void)
{
	union REGS r;

	Video_SetPlaneBitmask(15); // select all planes

	// clear whole VRAM area
__asm
{
	push eax
	push ecx
	push edi
	xor eax,eax
	mov edi,VGA_FRAMEBUFFER
	mov ecx,65536 / 4
	rep stosd
	pop edi
	pop ecx
	pop eax
};

	// set old video mode
	r.w.ax = origVideoMode;
	int386(0x10, &r, &r);
}

void Video_SetPal(uint8_t *pal) // 768 bytes
{
__asm
{
	push eax
	push edx
	push ecx
	push esi
	mov dx,0x3C8
	xor eax,eax
	out dx,al
	inc edx
	mov esi,pal
	mov ecx,3*VGA_PAL_NUM
	rep outsb
	pop esi
	pop ecx
	pop edx
	pop eax
};
}

void Video_GetPal(uint8_t *pal) // 768 bytes
{
__asm
{
	push eax
	push edx
	push ecx
	push edi
	mov dx,0x3C7
	xor eax,eax
	out dx,al
	add edx,2
	mov edi,pal
	mov ecx,3*VGA_PAL_NUM
	rep insb
	pop edi
	pop ecx
	pop edx
	pop eax
};
}

void Video_SetPalColor(uint8_t index, uint8_t r, uint8_t g, uint8_t b)
{
__asm
{
	push eax
	push edx
	mov dx,0x3C8
	mov al,index
	out dx,al
	inc edx
	mov al,r
	out dx,al
	mov al,g
	out dx,al
	mov al,b
	out dx,al
	pop edx
	pop eax
};
}

void Video_GetPalColor(uint8_t index, uint8_t *r, uint8_t *g, uint8_t *b)
{
__asm
{
	push eax
	push edx
	mov dx,0x3C7
	mov al,index
	out dx,al
	add edx,2
	in al,dx
	mov byte ptr [r],al
	in al,dx
	mov byte ptr [g],al
	in al,dx
	mov byte ptr [b],al
	pop edx
	pop eax
};
}

void Video_SetMouseColor(uint8_t r, uint8_t g, uint8_t b)
{
	if (vblankPitTimerEnabled)
	{
		mouseR = r;
		mouseG = g;
		mouseB = b;
		updateMousePalette = true; // will be done in PIT vblank interrupt
	}
	else
	{
__asm
{
	push eax
	push ebx
	push ecx
	push edx

	mov dx,0x3C8
	mov al,16
	out dx,al
	inc edx
	mov ecx,8
L0:	mov al,r
	out dx,al
	mov al,g
	out dx,al
	mov al,b
	out dx,al
	dec ecx
	jnz L0

	pop edx
	pop ecx
	pop ebx
	pop eax
};
	}
}
