#ifndef __VIDEO_H__
#define __VIDEO_H__

#include <stdint.h>
#include <stdbool.h>

#define SCREEN_W 320
#define SCREEN_H (240-1)
#define VGA_FRAMEBUFFER 0xA0000

// accurate nominal values

// 320x240 (640x480) 60Hz
#define VBLANK_HZ 59.940476190476

#define AMIGA_PAL_VBLANK_HZ 49.920409283472

// excluding mouse color (pal entry 8..15)
#define PAL_NUM_COLORS 8 

enum
{
	PAL_BACKGRD	= 0,
	PAL_BORDER	= 1,
	PAL_GENBKG	= 2,
	PAL_GENBKG2	= 3,
	PAL_QADSCP	= 4,
	PAL_PATCURSOR	= 5,
	PAL_GENTXT	= 6,
	PAL_PATTXT	= 7
};

typedef struct pal_t
{
	uint8_t r, g, b;
} pal_t;

extern const pal_t origTrackerPal[PAL_NUM_COLORS];

bool Video_IsColorVGA(void);
void Video_SetPlaneBitmask(uint8_t bitmask);
void Video_SelectWritePlane(uint8_t plane);
void Video_SelectReadPlane(uint8_t plane);
void Video_SetBackgroundColor(uint8_t r, uint8_t g, uint8_t b);
void Video_RestoreBackgroundColor(void);
void Video_GetVideoMode(void);
void Video_SetNewVideoMode(void);
void Video_ResetVideoMode(void);
void Video_ResetTrackerPalette(void);
void Video_SetPal(uint8_t *pal); // 3*256 bytes
void Video_GetPal(uint8_t *pal); // 3*256 bytes
void Video_SetPalColor(uint8_t index, uint8_t r, uint8_t g, uint8_t b);
void Video_GetPalColor(uint8_t index, uint8_t *r, uint8_t *g, uint8_t *b);
void Video_SetMouseColor(uint8_t r, uint8_t g, uint8_t b);

extern volatile bool vblankPitTimerEnabled, updateMousePalette;
extern uint8_t mouseR, mouseG, mouseB;
extern uint16_t vblankPitTicks;

// avideo.asm
extern void Video_WaitForVblank(void);
extern void Video_MeasureVblankPitTicks(void);

#endif
