#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <stdlib.h>
#include <conio.h>
#include "audio/replayer.h"
#include "hardware/video.h"
#include "gfx/font.h"
#include "gfx/drawcode.h"
#include "ui.h"
#include "pattern.h"
#include "asm.h"

#define TOP_ROWS 6
#define BOTTOM_ROWS 6
#define PAT_CURSOR_Y 192

static bool cursorWritten = false;

static const uint8_t cursorXPos[6*4] =
{
	 4,  7,  8,  9, 10, 11, // channel #1
	13, 16, 17, 18, 19, 20, // channel #2
	22, 25, 26, 27, 28, 29, // channel #3
	31, 34, 35, 36, 37, 38  // channel #4
};

static const bool cursorIsAtNote[6*4] =
{
	true, false, false, false, false, false, // channel #1
	true, false, false, false, false, false, // channel #2
	true, false, false, false, false, false, // channel #3
	true, false, false, false, false, false  // channel #4
};

void Pattern_ClearCache(void)
{
	ASM_FastFill32(oldPatternRows, 0xFFFFFFFF, VISIBLE_ROWS*4*4);
}

void Pattern_RedrawCursor(void)
{
	cursorWritten = false;
	Pattern_SetCursorPos(ui.patt.cursorPos);
}

void Pattern_SetCursorPos(uint8_t newCursorPos)
{
	uint8_t oldCursorPos;

	if (newCursorPos > 23)
		newCursorPos = 23;
		
	oldCursorPos = ui.patt.cursorPos;
	ui.patt.cursorPos = newCursorPos;

	if (ui.samplerScreenShown)
		return;

	// erase old cursor & draw new cursor
	__asm
	{
		push eax
		push ebx
		push ecx
		push edx
		push esi
		push edi
		
		mov dx,0x03C4 // VGA reg
		mov al,0x02 // VGA index 2, set plane mask
		xor ebx,ebx
		mov cx,0xFFFF

		// check if we need to erase old cursor
		cmp byte ptr [cursorWritten],0
		je drawNewCursor
		movzx esi,byte ptr oldCursorPos
		movzx edi,byte ptr newCursorPos
		cmp esi,edi
		je drawNewCursor

		// ----------------------------
		// erase old cursor
		// ----------------------------
	
		movzx edi,byte ptr [cursorXPos+esi]
		add edi,VGA_FRAMEBUFFER+(PAT_CURSOR_Y*(SCREEN_W/8))	
		cmp byte ptr [cursorIsAtNote+esi],1
		je cursorWasAtNote

		mov ah,0x05
		out dx,ax
		mov [edi+(0*(SCREEN_W/8))],bl
		mov [edi+(1*(SCREEN_W/8))],bl
		mov ah,0x03
		out dx,ax
		mov [edi+(0*(SCREEN_W/8))],cl
		mov [edi+(1*(SCREEN_W/8))],cl
		jmp drawNewCursor

	cursorWasAtNote:	
		mov ah,0x05
		out dx,ax
		mov [edi+(0*(SCREEN_W/8))+0],bx
		mov [edi+(0*(SCREEN_W/8))+2],bl
		mov [edi+(1*(SCREEN_W/8))+0],bx
		mov [edi+(1*(SCREEN_W/8))+2],bl	
		mov ah,0x03
		out dx,ax
		mov [edi+(0*(SCREEN_W/8))+0],cx
		mov [edi+(0*(SCREEN_W/8))+2],cl
		mov [edi+(1*(SCREEN_W/8))+0],cx
		mov [edi+(1*(SCREEN_W/8))+2],cl
		// ----------------------------
		// ----------------------------

	drawNewCursor:
		movzx esi,byte ptr newCursorPos
		movzx edi,byte ptr [cursorXPos+esi]
		add edi,VGA_FRAMEBUFFER+(PAT_CURSOR_Y*(SCREEN_W/8))	
		cmp byte ptr [cursorIsAtNote+esi],1
		je cursorAtNote
		
		mov ah,0x03
		out dx,ax
		mov [edi+(0*(SCREEN_W/8))],bl
		mov [edi+(1*(SCREEN_W/8))],bl	
		mov ah,0x05
		out dx,ax
		mov [edi+(0*(SCREEN_W/8))],cl
		mov [edi+(1*(SCREEN_W/8))],cl	
		jmp done
	
	cursorAtNote:	
		mov ah,0x03
		out dx,ax
		mov [edi+(0*(SCREEN_W/8))+0],bx
		mov [edi+(0*(SCREEN_W/8))+2],bl
		mov [edi+(1*(SCREEN_W/8))+0],bx
		mov [edi+(1*(SCREEN_W/8))+2],bl
		mov ah,0x05
		out dx,ax
		mov [edi+(0*(SCREEN_W/8))+0],cx
		mov [edi+(0*(SCREEN_W/8))+2],cl
		mov [edi+(1*(SCREEN_W/8))+0],cx
		mov [edi+(1*(SCREEN_W/8))+2],cl

	done:	mov byte ptr [cursorWritten],1
		pop edi
		pop esi
		pop edx
		pop ecx
		pop ebx
		pop eax
	};
}

void Pattern_Draw(int32_t pattern)
{
	int32_t row, currRow, rows, endRow, startRow;

	currRow = module.row;
	startRow = 0;

	row = currRow - TOP_ROWS;
	if (row < 0)
		row = 0;

	endRow = row + VISIBLE_ROWS;

	Video_SetPlaneBitmask(7);

	// clear top/bottom outside rows
	if (currRow < TOP_ROWS)
	{
		startRow = TOP_ROWS - currRow;
		endRow -= startRow;

		Pattern_ClearRows(0, startRow);
	}
	else if (currRow > 63-BOTTOM_ROWS)
	{
		rows = currRow-(63-BOTTOM_ROWS);
		endRow -= rows;	

		Pattern_ClearRows(TOP_ROWS+(64-currRow), rows);
	}

	Pattern_DrawRows(endRow, row, startRow, &patternData[(pattern << 8) + (row << 2)]);
}
