#ifndef __UI_H__
#define __UI_H__

#include <stdint.h>
#include <stdbool.h>
#include "hardware/video.h"

#define ERR_DURATION (uint8_t)(VBLANK_HZ/1.1)

enum // for UI_SetStatusText
{
	NO_CARRY = false,
	DO_CARRY = true
};

typedef struct
{
	struct
	{
		uint8_t cursorPos;
	} patt;

	volatile bool redrawPatternFlag, redrawPositionFlag, redrawBPMFlag, redrawTempoModeFlag, redrawModSizeFlag;

	bool markBlock;
	bool askBoxShown, samplerScreenShown, diskOpShown, editOpShown, setupShown;
	uint8_t f6Pos, f7Pos, f8Pos, f9Pos, f10Pos, editSkip, screen;
	int32_t currSample;
} ui_t;

extern ui_t ui;
extern volatile bool trackerRunning; // main.c
extern volatile bool showRasterBars;
extern const char HexTab[16];
extern char endMsg[128];

extern void UI_AmigaVblank(void);

void UI_SetPrevStatusMessage(void);
void UI_Debug(const char *text); // safe to use in interrupts/threads
void UI_Debugf(const char *fmt, ...); // safe to use in interrupts/threads
void UI_PrintQuitMessage(void);
void UI_Error(const char *text);
void UI_ShowTemporaryStatusText(const char *text, uint16_t duration);
void UI_HandleUpdates(void);
bool AskBox(const char *text);
bool UI_CreateScopeTable(void);
void UI_FreeScopeTable(void);
void UI_DrawScopes(void);
void UI_DrawStatusText(void);
void UI_SetStatusText(const char *text, bool carry);
void UI_StatusAllRight(void);
void UI_DrawModSize(void);
void UI_DrawTempoMode(void);
void UI_DrawEditPattern(void);
void UI_DrawPosInfo(void);
void UI_DrawSongLength(void);
void UI_DrawSampleInfo(void);
void UI_ResetPlaybackTimer(void);
void UI_DrawPlaybackTimer(void);
void UI_DrawSongTempo(void);
void UI_DrawSongName(void);
void UI_DrawSampleName(void);
void UI_DrawEditSkip(void);
void UI_HandleKeyPress(uint8_t scancode);
void UI_DrawTopScreen(void);
void UI_DrawMiddleScreen(void);
void UI_DrawBottomScreen(void);
void UI_HandleMouseClick(void);

// aui.asm
extern void UI_AmigaVblankHandler(void);
extern void UI_ResetAmigaVblankHandlerFrac(void);

#endif
